#! /bin/env python

import subprocess
from pathlib import Path
from .logger import log

def to_yuv(file: Path):
    output = file.with_suffix('.yuv')
    cmd = f'ffmpeg -y -loglevel error -i {file.as_posix()} -f rawvideo -vf format=yuv420p10le {output.as_posix()}'
    subprocess.run(cmd, shell=True)
    return output


def vvc_concat(temp: Path, output: Path):
    encode_files = sorted((temp / 'encode').iterdir())
    bitstreams = [x.as_posix() for x in encode_files]
    bitstreams = ' '.join(bitstreams)
    cmd = f'vvc_concat  {bitstreams} {output.as_posix()}'

    output = subprocess.run(cmd, shell=True)




#!/usr/bin/env python3
import json

def conf(args):
    """Creation and reading of config files with saved settings"""
    if args.config:
        if args.config.exists():
            with open(args.config) as f:
                c: dict = dict(json.load(f))
                args.__dict__.update(c)

        else:
            with open(args.config, 'w') as f:
                c = dict()
                c['video_params'] = args.video_params
                c['encoder'] = args.encoder
                c['ffmpeg'] = args.ffmpeg
                c['audio_params'] = args.audio_params
                json.dump(c, f)




#!/bin/env python

import json
import re
import statistics
import subprocess
import sys
from pathlib import Path
from subprocess import PIPE


import cv2
import numpy as np

def terminate():
    sys.exit(1)


def process_inputs(inputs):
    # Check input file for being valid
    if not inputs:
        print('No input file')
        terminate()

    if inputs[0].is_dir():
        inputs = [x for x in inputs[0].iterdir() if x.suffix in (".mkv", ".mp4", ".mov", ".avi", ".flv", ".m2ts")]

    valid = np.array([i.exists() for i in inputs])

    if not all(valid):
        print(f'File(s) do not exist: {", ".join([str(inputs[i]) for i in np.where(not valid)[0]])}')
        terminate()

    return inputs


def get_cq(command):
    """
    Return cq values from command
    :param command: string with commands for encoder
    :return: list with frame numbers of keyframes

    """
    matches = re.findall(r"--cq-level= *([^ ]+?) ", command)
    return int(matches[-1])


def man_q(command: str, q: int):
    """Return command with new cq value"""

    if 'aomenc' in command or 'vpxenc' in command:
        mt = '--cq-level='
        cmd = command[:command.find(mt) + 11] + str(q) + ' ' + command[command.find(mt) + 13:]

    elif 'x265' in command:
        mt = '--crf'
        cmd = command[:command.find(mt) + 6] + str(q) + ' ' +  command[command.find(mt) + 9:]

    elif 'rav1e' in command:
        mt = '--quantizer'
        cmd = command[:command.find(mt) + 11] + ' ' + str(q) + ' ' +  command[command.find(mt) + 15:]

    elif 'SvtAv1EncApp' in command:
        mt = '--qp'
        cmd = command[:command.find(mt) + 4] + ' ' + str(q) + ' ' +  command[command.find(mt) + 7:]
    return cmd


def frame_probe_cv2(source: Path):
    video = cv2.VideoCapture(source.as_posix())
    total = int(video.get(cv2.CAP_PROP_FRAME_COUNT))
    return total


def get_brightness(video):
    """Getting average brightness value for single video."""
    brightness = []
    cap = cv2.VideoCapture(video)
    try:
        while True:
            # Capture frame-by-frame
            _, frame = cap.read()

            # Our operations on the frame come here
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

            # Display the resulting frame
            mean = cv2.mean(gray)
            brightness.append(mean[0])
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break
    except cv2.error:
        pass

    # When everything done, release the capture
    cap.release()
    brig_geom = round(statistics.geometric_mean([x + 1 for x in brightness]), 1)

    return brig_geom



#!/bin/env python

import json
import subprocess
from pathlib import Path
from subprocess import PIPE, STDOUT
from .utils import terminate
from .logger import log
from threading import Lock
import re
import sys


def frame_probe(source: Path):
    """Get frame count."""
    cmd = ["ffmpeg", "-hide_banner", "-i", source.as_posix(), "-map", "0:v:0", "-f", "null", "-"]
    r = subprocess.run(cmd, stdout=PIPE, stderr=PIPE)
    matches = re.findall(r"frame=\s*([0-9]+)\s", r.stderr.decode("utf-8") + r.stdout.decode("utf-8"))
    return int(matches[-1])


def get_keyframes(file: Path):
    """
    Read file info and return list of all keyframes

    :param file: Path for input file
    :return: list with frame numbers of keyframes
    """

    keyframes = []

    ff = ["ffmpeg", "-hide_banner", "-i", file.as_posix(),
    "-vf", r"select=eq(pict_type\,PICT_TYPE_I)",
    "-f", "null", "-loglevel", "debug", "-"]

    pipe = subprocess.Popen(ff, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    while True:
        line = pipe.stdout.readline().strip().decode("utf-8")

        if len(line) == 0 and pipe.poll() is not None:
            break

        match = re.search(r"n:([0-9]+)\.[0-9]+ pts:.+key:1", line)
        if match:
            keyframe = int(match.group(1))
            keyframes.append(keyframe)

    return keyframes


doneFileLock = Lock()
def write_progress_file(file, chunk, frames):
    doneFileLock.acquire()
    with file.open() as f:
        d = json.load(f)
    d['done'][chunk.name] = frames
    with file.open('w') as f:
        json.dump(d, f)


def frame_check(source: Path, encoded: Path, temp, check):
    """Checking is source and encoded video frame count match."""
    try:
        status_file = Path(temp / 'done.json')

        if check:
            s1 = frame_probe(source)
            write_progress_file(status_file, source, s1)
        else:
            s1, s2 = [frame_probe(i) for i in (source, encoded)]
            if s1 == s2:
                write_progress_file(status_file, source, s1)
            else:
                print(f'Frame Count Differ for Source {source.name}: {s2}/{s1}')

    except IndexError:
        print('Encoding failed, check validity of your encoding settings/commands and start again')
        terminate()
    except Exception as e:
        _, _, exc_tb = sys.exc_info()
        print(f'\nError frame_check: {e}\nAt line: {exc_tb.tb_lineno}\n')
    finally:
        if doneFileLock.locked():
            doneFileLock.release()


def concatenate_video(temp, output, encoder):
    """With FFMPEG concatenate encoded segments into final file."""

    log('Concatenating\n')

    with open(f'{temp / "concat" }', 'w') as f:

        encode_files = sorted((temp / 'encode').iterdir())
        # Replace all the ' with '/'' so ffmpeg can read the path correctly
        f.writelines("file '" + str(file.absolute()).replace('\'','\'\\\'\'') + "'\n" for file in encode_files)

    # Add the audio file if one was extracted from the input
    audio_file = temp / "audio.mkv"
    if audio_file.exists():
        audio = f'-i {audio_file} -c:a copy -map 1'
    else:
        audio = ''

    if encoder == 'x265':

        cmd = f' ffmpeg -y -fflags +genpts  -hide_banner -loglevel error -f concat -safe 0 -i {temp / "concat"} ' \
            f'{audio} -c copy -movflags frag_keyframe+empty_moov -map 0  -f mp4 - | ffmpeg -y -hide_banner -loglevel error -i - -c copy {output} '
        concat = subprocess.run(cmd, shell=True, stdout=PIPE, stderr=STDOUT).stdout

    else:
        cmd = f' ffmpeg -y -hide_banner -loglevel error -f concat -safe 0 -i {temp / "concat"} ' \
            f'{audio} -c copy -map 0  -y "{output}"'

        concat = subprocess.run(cmd, shell=True, stdout=PIPE, stderr=STDOUT).stdout


    if len(concat) > 0:
        log(concat.decode())
        print(concat.decode())
        raise Exception


def extract_audio(input_vid: Path, temp, audio_params):
    """Extracting audio from source, transcoding if needed."""
    log(f'Audio processing\nParams: {audio_params}\n')
    audio_file = temp / 'audio.mkv'

    # Checking is source have audio track
    check = fr' ffmpeg -y -hide_banner -loglevel error -ss 0 -i "{input_vid}" -t 0 -vn -c:a copy -f null -'
    is_audio_here = len(subprocess.run(check, shell=True, stdout=PIPE, stderr=STDOUT).stdout) == 0

    # If source have audio track - process it
    if is_audio_here:
        cmd = f'ffmpeg -y -hide_banner -loglevel error -i "{input_vid}" -vn ' \
              f'{audio_params} {audio_file}'
        subprocess.run(cmd, shell=True)



#!/bin/env python

import os
import sys
import shutil
import atexit
from distutils.spawn import find_executable
from pathlib import Path

from psutil import virtual_memory

from .utils import terminate
from .compose import get_default_params_for_encoder


def set_vmaf(args):

    if args.vmaf_path:
        if not Path(args.vmaf_path).exists():
            print(f'No such model: {Path(args.vmaf_path).as_posix()}')
            terminate()

    if args.vmaf_steps < 4:
        print('Target vmaf require more than 3 probes/steps')
        terminate()


    defaul_ranges = {'svt_av1': (20, 40), 'rav1e': (70, 150), 'aom': (25, 50), 'vpx': (25, 50),'x265': (20, 40), 'vvc': (20, 50)}

    if args.min_q is None or args.max_q is None:
        args.min_q, args.max_q = defaul_ranges.get(args.encoder)


def check_exes(args):
    if not find_executable('ffmpeg'):
        print('No ffmpeg')
        terminate()
    encoders = {'svt_av1': 'SvtAv1EncApp', 'rav1e': 'rav1e', 'aom': 'aomenc', 'vpx': 'vpxenc','x265': 'x265', 'vvc': 'vvc_encoder'}


    # Check if encoder executable is reachable
    if args.encoder in encoders:
        enc = encoders.get(args.encoder)

        if not find_executable(enc):
            print(f'Encoder {enc} not found')
            terminate()
    else:
        print(f'Not valid encoder {args.encoder}\nValid encoders: "aom rav1e", "svt_av1", "vpx", "x265" ')
        terminate()

    if args.encoder == 'vvc':
        if not find_executable('vvc_concat'):
            print('vvc concatenation executabe "vvc_concat" not found')
            terminate()


def startup_check(args):

    encoders_default_passes = {'svt_av1': 1, 'rav1e': 1, 'aom': 2, 'vpx': 2,'x265': 1, 'vvc':1 }


    if sys.version_info < (3, 6):
        print('Python 3.6+ required')
        sys.exit()
    if sys.platform == 'linux':
        def restore_term():
            os.system("stty sane")
        atexit.register(restore_term)

    check_exes(args)

    set_vmaf(args)

    if args.reuse_first_pass and args.encoder != 'aom' and args.split_method != 'aom_keyframes':
        print('Reusing the first pass is only supported with the aom encoder and aom_keyframes split method.')
        terminate()


    if args.encoder == 'vvc' and not args.vvc_conf:
        print('Conf file for vvc required')
        terminate()

    # No check because vvc
    if args.encoder == 'vvc':
        args.no_check = True

    if args.passes is None:
        args.passes = encoders_default_passes.get(args.encoder)


    if args.video_params is None and args.encoder == 'vvc':
        print('VVC require:\n',
        ' -wdt X - video width\n',
        ' -hgt X - video height\n',
        ' -fr X  - framerate\n',
        ' -q X   - quantizer\n',
        'Example: -wdt 640 -hgt 360 -fr 23.98 -q 30 '
        )
        terminate()

    if args.video_params is None:
        args.video_params = get_default_params_for_encoder(args.encoder)



    args.pix_format = f'-strict -1 -pix_fmt {args.pix_format}'
    args.ffmpeg_pipe = f' {args.ffmpeg} {args.pix_format} -f yuv4mpegpipe - |'


def determine_resources(encoder, workers):
    """Returns number of workers that machine can handle with selected encoder."""

    # If set by user, skip
    if workers != 0:
        return workers

    cpu = os.cpu_count()
    ram = round(virtual_memory().total / 2 ** 30)

    if encoder in ('aom', 'rav1e', 'vpx'):
        workers =  round(min(cpu / 2, ram / 1.5))

    elif encoder in ('svt_av1', 'x265'):
        workers =  round(min(cpu, ram)) // 8

    elif encoder in ('vvc'):
        workers = round(min(cpu, ram))

    # fix if workers round up to 0
    if workers == 0:
        workers = 1

    return workers


def setup(temp: Path, resume):
    """Creating temporally folders when needed."""
    # Make temporal directories, and remove them if already presented
    if not resume:
        if temp.is_dir():
            shutil.rmtree(temp)

    (temp / 'split').mkdir(parents=True, exist_ok=True)
    (temp / 'encode').mkdir(exist_ok=True)


def outputs_filenames(inp: Path, out:Path, encoder):
    suffix = '.mkv'
    if out:
        return out.with_suffix(suffix)
    else:
        return Path(f'{inp.stem}_av1{suffix}')



#!/usr/bin/env python3

from Av1an import *


class Av1an:
    """Av1an - Python framework for AV1, VP9, VP8 encoding"""
    def __init__(self):
        self.args = arg_parsing()

    def main_thread(self):
        """Main."""
        startup_check(self.args)
        main_queue(self.args)


def main():
    Av1an().main_thread()


if __name__ == '__main__':
    main()



#!/bin/env python

from .utils import terminate
from .ffmpeg import frame_probe
from .vmaf import call_vmaf, read_vmaf_json
from scipy import interpolate
from pathlib import Path
import subprocess
import numpy as np
from .logger import log
from matplotlib import pyplot as plt
import matplotlib
import sys
from math import isnan
import os
from .bar import make_pipes

def x264_probes(video: Path, ffmpeg: str):
    cmd = f' ffmpeg -y -hide_banner -loglevel error -i {video.as_posix()} ' \
                  f'-r 4 -an {ffmpeg} -c:v libx264 -crf 0 {video.with_suffix(".mp4")}'
    subprocess.run(cmd, shell=True)


def gen_probes_names(probe, q):
    """Make name of vmaf probe
    """
    return probe.with_name(f'v_{q}{probe.stem}').with_suffix('.ivf')


def probe_cmd(probe, q, ffmpeg_pipe, encoder):
    """Generate and return commands for probes at set Q values
    """
    pipe = f'ffmpeg -y -hide_banner -loglevel error -i {probe} {ffmpeg_pipe}'

    if encoder == 'aom':
        params = " aomenc  -q --passes=1 --threads=8 --end-usage=q --cpu-used=6 --cq-level="
        cmd = f'{pipe} {params}{q} -o {probe.with_name(f"v_{q}{probe.stem}")}.ivf - '

    elif encoder == 'x265':
        params = "x265  --log-level 0  --no-progress --y4m --preset faster --crf "
        cmd = f'{pipe} {params}{q} -o {probe.with_name(f"v_{q}{probe.stem}")}.ivf - '

    elif encoder == 'rav1e':
        params = "rav1e - -q -s 10 --tiles 8 --quantizer "
        cmd = f'{pipe} {params}{q} -o {probe.with_name(f"v_{q}{probe.stem}")}.ivf'

    elif encoder == 'vpx':
        params = "vpxenc --passes=1 --pass=1 --codec=vp9 --threads=4 --cpu-used=9 --end-usage=q --cq-level="
        cmd = f'{pipe} {params}{q} -o {probe.with_name(f"v_{q}{probe.stem}")}.ivf - '

    elif encoder == 'svt_av1':
        params = " SvtAv1EncApp -i stdin --preset 8 --rc 0 --qp "
        cmd = f'{pipe} {params}{q} -b {probe.with_name(f"v_{q}{probe.stem}")}.ivf'

    return cmd


def get_target_q(scores, vmaf_target):
    x = [x[1] for x in sorted(scores)]
    y = [float(x[0]) for x in sorted(scores)]
    f = interpolate.interp1d(x, y, kind='cubic')
    xnew = np.linspace(min(x), max(x), max(x) - min(x))
    tl = list(zip(xnew, f(xnew)))
    q = min(tl, key=lambda x: abs(x[1] - vmaf_target))

    return int(q[0]), round(q[1],3)


def interpolate_data(vmaf_cq: list, vmaf_target):
    x = [x[1] for x in sorted(vmaf_cq)]
    y = [float(x[0]) for x in sorted(vmaf_cq)]

    # Interpolate data
    f = interpolate.interp1d(x, y, kind='cubic')
    xnew = np.linspace(min(x), max(x), max(x) - min(x))

    # Getting value closest to target
    tl = list(zip(xnew, f(xnew)))
    vmaf_target_cq = min(tl, key=lambda x: abs(x[1] - vmaf_target))
    return vmaf_target_cq, tl, f, xnew


def plot_probes(args, vmaf_cq, vmaf_target, probe, frames):
    # Saving plot of vmaf calculation

    x = [x[1] for x in sorted(vmaf_cq)]
    y = [float(x[0]) for x in sorted(vmaf_cq)]

    cq, tl, f, xnew = interpolate_data(vmaf_cq, args.vmaf_target)
    matplotlib.use('agg')
    plt.ioff()
    plt.plot(xnew, f(xnew), color='tab:blue', alpha=1)
    plt.plot(x, y, 'p', color='tab:green', alpha=1)
    plt.plot(cq[0], cq[1], 'o', color='red', alpha=1)
    plt.grid(True)
    plt.xlim(args.min_q, args.max_q)
    vmafs = [int(x[1]) for x in tl if isinstance(x[1], float) and not isnan(x[1])]
    plt.ylim(min(vmafs), max(vmafs) + 1)
    plt.ylabel('VMAF')
    plt.title(f'Chunk: {probe.stem}, Frames: {frames}')
    plt.tight_layout()
    temp = args.temp / probe.stem
    plt.savefig(f'{temp}.png', dpi=200, format='png')
    plt.close()


def vmaf_probe(probe, q, args):

    cmd = probe_cmd(probe, q, args.ffmpeg_pipe, args.encoder)
    make_pipes(cmd).wait()

    file = call_vmaf(probe, gen_probes_names(probe, q), args.n_threads, args.vmaf_path)
    score = read_vmaf_json(file, 25)

    return score


def early_skips(probe, source, frames,args):

    cq = [args.max_q, args.min_q]
    scores = []
    for i in (0, 1):

        score = vmaf_probe(probe, cq[i], args)
        scores.append((score, cq[i]))
        # Early Skip on big CQ
        if i == 0 and round(score) > args.vmaf_target:
            log(f"File: {source.stem}, Fr: {frames}\n" \
            f"Q: {sorted([x[1] for x in scores])}, Early Skip High CQ\n" \
            f"Vmaf: {sorted([x[0] for x in scores], reverse=True)}\n" \
            f"Target Q: {args.max_q} Vmaf: {score}\n\n")

            return True, args.max_q

        # Early Skip on small CQ
        if i == 1 and round(score) < args.vmaf_target:
            log(f"File: {source.stem}, Fr: {frames}\n" \
                f"Q: {sorted([x[1] for x in scores])}, Early Skip Low CQ\n" \
                f"Vmaf: {sorted([x[0] for x in scores], reverse=True)}\n" \
                f"Target Q: {args.min_q} Vmaf: {score}\n\n")

            return True, args.min_q

    return False, scores


def get_closest(q_list, q, positive=True):
    """Returns closest value from the list, ascending or descending
    """
    if positive:
        q_list = [x for x in q_list if x > q]
    else:
        q_list = [x for x in q_list if x < q]

    return min(q_list, key=lambda x:abs(x-q))


def target_vmaf_search(probe, source, frames, args):

    vmaf_cq = []
    q_list = [args.min_q, args.max_q]
    score = 0
    last_q = args.min_q
    next_q = args.max_q
    for _ in range(args.vmaf_steps - 2 ):

        new_point= (last_q + next_q) // 2
        last_q = new_point

        q_list.append(new_point)
        score = vmaf_probe(probe, new_point, args)
        next_q = get_closest(q_list, last_q, positive=score >= args.vmaf_target)
        vmaf_cq.append((score, new_point))

    return vmaf_cq


def target_vmaf(source, args):

    frames = frame_probe(source)
    probe = source.with_suffix(".mp4")
    vmaf_cq = []

    try:
        x264_probes(source, args.ffmpeg)

        skips, scores = early_skips(probe, source, frames, args)
        if skips:
            return scores
        else:
            vmaf_cq.extend(scores)

        scores = target_vmaf_search(probe, source, frames, args)

        vmaf_cq.extend(scores)

        q, q_vmaf = get_target_q(vmaf_cq, args.vmaf_target )

        log(f'File: {source.stem}, Fr: {frames}\n' \
            f'Q: {sorted([x[1] for x in vmaf_cq])}\n' \
            f'Vmaf: {sorted([x[0] for x in vmaf_cq], reverse=True)}\n' \
            f'Target Q: {q} Vmaf: {q_vmaf}\n\n')

        if args.vmaf_plots:
            plot_probes(args, vmaf_cq, args.vmaf_target, probe, frames)

        return q

    except Exception as e:
        _, _, exc_tb = sys.exc_info()
        print(f'Error in vmaf_target {e} \nAt line {exc_tb.tb_lineno}')
        terminate()



#!/usr/bin/env python3
import time
import json
import os
from pathlib import Path
from .target_vmaf import target_vmaf
from .boost import boosting
from .utils import  frame_probe_cv2, terminate, process_inputs
from .fp_reuse import remove_first_pass_from_commands
from .utils import man_q
from .logger import log
from .bar import Manager, tqdm_bar
from .setup import (startup_check, determine_resources, outputs_filenames, setup)
from .aom_kf import AOM_KEYFRAMES_DEFAULT_PARAMS
import sys
import concurrent
import concurrent.futures
from .logger import log, set_log
from .config  import conf
from .compose import compose_encoding_queue, get_video_queue
from .ffmpeg import concatenate_video, extract_audio, frame_probe, frame_check
from .fp_reuse import segment_first_pass
from .split import extra_splits, segment, split_routine
from .vmaf import plot_vmaf
import shutil
from .vvc import to_yuv, vvc_concat


def video_encoding(args):
        """Encoding video on local machine."""

        args.output_file = outputs_filenames(args.input, args.output_file, args.encoder )

        if args.resume and (args.temp / 'done.json').exists():
            set_log(args.logging, args.temp)
        else:
            setup(args.temp, args.resume)
            set_log(args.logging, args.temp)

            # inherit video params from aom encode unless we are using a different encoder, then use defaults
            aom_keyframes_params = args.video_params if (args.encoder == 'aom') else AOM_KEYFRAMES_DEFAULT_PARAMS
            framenums = split_routine(args, aom_keyframes_params)

            if args.extra_split:
                framenums = extra_splits(args.input, framenums, args.extra_split)

            if args.reuse_first_pass:
                segment_first_pass(args.temp, framenums)

            segment(args.input, args.temp, framenums)
            extract_audio(args.input, args.temp,  args.audio_params)

        chunk = get_video_queue(args.temp,  args.resume)

        # Make encode queue
        commands = compose_encoding_queue(chunk, args)

        args.workers = determine_resources(args.encoder, args.workers)

        encoding_loop(args, commands)

        try:
            if args.encoder == 'vvc':
                vvc_concat(args.temp, args.output_file.with_suffix('.h266'))
            else:
                concatenate_video(args.temp, args.output_file, args.encoder )

        except Exception as e:
            _, _, exc_tb = sys.exc_info()
            print(f'Concatenation failed, FFmpeg error\nAt line: {exc_tb.tb_lineno}\nError:{str(e)}')
            log(f'Concatenation failed, aborting, error: {e}\n')
            terminate()

        if args.vmaf or args.vmaf_plots:
            plot_vmaf(args.input, args.output_file, args.vmaf_path)
        # Delete temp folders
        if not args.keep:
            shutil.rmtree(args.temp)


def main_queue(args):
    # Todo: Redo Queue
    try:
        conf(args)
        tm = time.time()

        args.queue = process_inputs(args.input)

        for file in args.queue:
            tm = time.time()
            args.input = file

            if len(args.queue) > 1:
                print(f'Encoding: {file}')
                args.output_file = None

            video_encoding(args)
            print(f'Finished: {round(time.time() - tm, 1)}s\n')
    except KeyboardInterrupt:
        print('Encoding stopped')
        sys.exit()


def encoding_loop(args, commands):
        """Creating process pool for encoders, creating progress bar."""
        try:
            enc_path = args.temp / 'split'
            done_path =args.temp / 'done.json'

            if args.resume and done_path.exists():
                log('Resuming...\n')

                with open(done_path) as f:
                    data = json.load(f)

                total = data['total']
                done = len(data['done'])
                initial = sum(data['done'].values())

                log(f'Resumed with {done} encoded clips done\n\n')
            else:
                initial = 0
                total = frame_probe_cv2(args.input)

                if total < 1:
                    total = frame_probe(args.input)

                d = {'total': total, 'done': {}}
                with open(done_path, 'w') as f:
                    json.dump(d, f)

            clips = len([x for x in enc_path.iterdir() if x.suffix == ".mkv"])
            args.workers = min(args.workers, clips)

            print(f'\rQueue: {clips} Workers: {args.workers} Passes: {args.passes}\n'
                  f'Params: {args.video_params.strip()}')

            with concurrent.futures.ThreadPoolExecutor(max_workers=args.workers) as executor:
                counter = Manager().Counter(total, initial)
                future_cmd = {executor.submit(encode, (cmd, counter, args)): cmd for cmd in commands}
                for future in concurrent.futures.as_completed(future_cmd):
                    future_cmd[future]
                    try:
                        future.result()
                    except Exception as exc:
                        _, _, exc_tb = sys.exc_info()
                        print(f'Encoding error {exc}\nAt line {exc_tb.tb_lineno}')
                        terminate()
        except KeyboardInterrupt:
            terminate()


def encode(commands):
    """Main encoding flow, appliying all scene based features

    :param args: Arguments object
    :param commands: composed commands for encoding
    :type commands: list
    """
    commands, counter, args  = commands
    try:
        st_time = time.time()
        source, target = Path(commands[-1][0]), Path(commands[-1][1])
        frame_probe_source = frame_probe(source)

        # Target Vmaf Mode
        if args.vmaf_target:
            tg_cq = target_vmaf(source, args)
            cm1 = man_q(commands[0], tg_cq, )

            if args.passes == 2:
                cm2 = man_q(commands[1], tg_cq)
                commands = (cm1, cm2) + commands[2:]
            else:
                commands = (cm1,) + commands[1:]

        # Boost
        if args.boost:
            commands = boosting(args.boost_limit, args.boost_range, source, commands, args.passes)

        # first pass reuse
        if args.reuse_first_pass:
            commands = remove_first_pass_from_commands(commands, args.passes)

        log(f'Enc: {source.name}, {frame_probe_source} fr\n\n')

        # Queue execution
        for i in commands[:-1]:
                if args.encoder == 'vvc':
                    log(f'Creating yuv for file {commands[1][0]}\n')
                    fl = to_yuv(commands[1][0])
                    log(f'Created yuv for file {commands[1][0]}\n')
                tqdm_bar(i, args.encoder, counter, frame_probe_source, args.passes)

                if args.encoder == 'vvc':
                    os.remove(fl)

        frame_check(source, target, args.temp, args.no_check)

        if args.encoder == 'vvc':
            frame_probe_fr = frame_probe(source)
        else:
             frame_probe_fr = frame_probe(target)

        enc_time = round(time.time() - st_time, 2)

        log(f'Done: {source.name} Fr: {frame_probe_fr}\n'
            f'Fps: {round(frame_probe_fr / enc_time, 4)} Time: {enc_time} sec.\n\n')

    except Exception as e:
        _, _, exc_tb = sys.exc_info()
        print(f'Error in encoding loop {e}\nAt line {exc_tb.tb_lineno}')



#!/bin/env python

import re
import subprocess
import sys
from collections import deque
from multiprocessing.managers import BaseManager
from subprocess import PIPE, STDOUT
from tqdm import tqdm

from .utils import terminate


def Manager():
    m = BaseManager()
    m.start()
    return m


class Counter():
    def __init__(self, total, initial):
        self.first_update = True
        self.initial = initial
        self.left = total - initial
        self.tqdm_bar = tqdm(total=self.left, initial=0, dynamic_ncols=True, unit="fr", leave=True, smoothing=0.2)

    def update(self, value):
        if self.first_update:
            self.tqdm_bar.reset(self.left)
            self.first_update = False
        self.tqdm_bar.update(value)


BaseManager.register('Counter', Counter)

def make_pipes(command):

    f, e = command.split('|')
    f, e = f.split(), e.split()
    ffmpeg_pipe = subprocess.Popen(f, stdout=PIPE, stderr=STDOUT)
    pipe = subprocess.Popen(e, stdin=ffmpeg_pipe.stdout, stdout=PIPE,
                            stderr=STDOUT,
                            universal_newlines=True)

    return pipe


def make_vvc_pipe(command):
    pipe = subprocess.Popen(command.split(), stdout=PIPE,
                            stderr=STDOUT,
                            universal_newlines=True)
    return pipe

def match_aom_vpx(line):
    if 'fatal' in line.lower():
        print('\n\nERROR IN ENCODING PROCESS\n\n', line)
        terminate()
    if 'Pass 2/2' in line or 'Pass 1/1' in line:
        return re.search(r"frame.*?\/([^ ]+?) ", line)


def match_rav1e(line):
    if 'error' in line.lower():
        print('\n\nERROR IN ENCODING PROCESS\n\n', line)
        terminate()
    return re.search(r"encoded.*? ([^ ]+?) ", line)


def match_vvc(line):
    return re.search(r"POC.*? ([^ ]+?)", line)


def process_encoding_pipe(pipe, encoder, counter):
    encoder_history = deque(maxlen=20)
    frame = 0
    pass_1_check = True
    skip_1_pass = False
    while True:
        line = pipe.stdout.readline().strip()

        if len(line) == 0 and pipe.poll() is not None:
            break

        if len(line) == 0:
            continue

        if encoder in ('aom', 'vpx'):
            match = match_aom_vpx(line)

        elif encoder == 'rav1e':
            match = match_rav1e(line)

        if encoder in  ('x265'):
            if not skip_1_pass and pass_1_check:
                if 'output file' in line:
                    if 'nul' in line.lower():
                        skip_1_pass = True
                    else:
                        pass_1_check = False
            if not skip_1_pass:
                match = re.search(r"^(\d+)", line)

        if encoder in ('vvc'):
            match = match_vvc(line)
            if match:
                counter.update(1)
                continue

        if match:
            new = int(match.group(1))
            if new > frame:
                counter.update(new - frame)
                frame = new

        if line:
            encoder_history.append(line)

    if pipe.returncode != 0 and pipe.returncode != -2:  # -2 is Ctrl+C for aom
        print(f"\nEncoder encountered an error: {pipe.returncode}")
        print('\n'.join(encoder_history))


def tqdm_bar(i, encoder, counter, frame_probe_source, passes):
    try:

        if encoder in 'vvc':
            pipe = make_vvc_pipe(i)
        else:
            pipe = make_pipes(i)

        if encoder in ('aom', 'vpx', 'rav1e','x265', 'vvc'):
            process_encoding_pipe(pipe, encoder, counter)

        if encoder == 'svt_av1':
            # SVT-AV1 developer: SVT-AV1 is special in the way it outputs to console
            pipe.wait()
            counter.update(frame_probe_source // passes)
    except Exception as e:
        _, _, exc_tb = sys.exc_info()
        print(f'Error at encode {e}\nAt line {exc_tb.tb_lineno}')